# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
PXE discovery support.
"""

import contextlib
import itertools

from oslo.config import cfg

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common import i18n
from ironic.common import states
from ironic.common import utils
from ironic.conductor import task_manager
from ironic.conductor import utils as conductor_utils
from ironic.db import api as dbapi
from ironic import objects
from ironic.openstack.common import log as logging


pxe_opts = [
    cfg.BoolOpt('discovery_enabled',
                default=False,
                help='Enable hardware discovery.'),
    cfg.BoolOpt('discovery_allow_search_by_mac',
                default=False,
                help='Whether to allow search of node by MAC '
                     '(required only for testing with pxe_ssh driver).'),
    cfg.BoolOpt('discovery_leave_maintenance',
                default=True,
                help='Whether to leave maintenance mode after discovery.'),
]


LOG = logging.getLogger(__name__)


CONF = cfg.CONF
CONF.register_opts(pxe_opts, group='pxe')


_LE = i18n._LE
_LI = i18n._LI
_LW = i18n._LW


@contextlib.contextmanager
def _lock_node(uuid):
    # We can't use task_manager, as it requires driver to be loaded
    db = dbapi.get_instance()
    node = db.reserve_node(CONF.host, uuid)
    try:
        yield node
    finally:
        db.release_node(CONF.host, node.id)


def continue_discovery(ctx, node_info):
    """Process data posted from the ramdisk."""
    if not CONF.pxe.discovery_enabled:
        LOG.error(_LE('Cannot process discovery data - disabled'))
        return

    if node_info.get('error'):
        LOG.error(_LE('Error happened during discovery: %s'),
                  node_info['error'])
        return

    keys = ('cpus', 'cpu_arch', 'memory_mb', 'local_gb', 'macs')
    missing = [key for key in keys if not node_info.get(key)]
    if missing:
        LOG.error(_LE('The following required parameters are missing: %s'),
                  missing)
        return

    valid_macs = [mac.lower() for mac in node_info['macs']
                  if utils.is_valid_mac(mac)]
    if valid_macs != node_info['macs']:
        LOG.warn(_LW('The following MACs were invalid in discovery data '
                     'for node with BMC %(ipmi_address)s and were '
                     'excluded: %(invalid)s'),
                 {'invalid': set(node_info['macs']) - set(valid_macs),
                  'ipmi_address': node_info.get('ipmi_address')})

    LOG.debug('Discovery data received from node with BMC '
              '%(ipmi_address)s: CPUs: %(cpus)s %(cpu_arch)s, '
              'memory %(memory_mb)s MiB, disk %(local_gb)s GiB, '
              'macs %(macs)s',
              dict((key, node_info.get(key))
                   for key in keys + ('ipmi_address',)))

    bmc_known = bool(node_info.get('ipmi_address'))
    if bmc_known:
        node = _get_node_by_ipmi_address(ctx, node_info['ipmi_address'])
        if node is None:
            LOG.error(_LE('Unable to find node with ipmi_address %s'),
                      node_info['ipmi_address'])
            return
    elif CONF.pxe.discovery_allow_search_by_mac:
        # In case of testing with vms and pxe_ssh driver
        LOG.warn(_LW('No BMC address provided, trying to use MAC '
                     'addresses for finding node'))
        node = _get_node_by_macs(ctx, valid_macs)
        if node is None:
            LOG.error(_LE('Unable to find node with macs %s'),
                      valid_macs)
            return
    else:
        LOG.error(_LE('No ipmi_address provided and searching by MAC is not '
                      'allowed'))
        return

    # Now properly lock node and proceed
    with _lock_node(node.uuid) as locked_node:
        _update_node(ctx, locked_node, node_info, valid_macs)

    # Discovery process will leave Node in 'power on' state in database,
    # as it's in maintenance mode and real power state is not synced.
    # NOTE(dtantsur): only works if the current conductor has required driver
    if node.driver in driver_factory.DriverFactory().names:
        LOG.debug('Trying to power off node %s after discovery', node.uuid)
        with task_manager.acquire(ctx, node.id) as task:
            conductor_utils.node_power_action(task, states.POWER_OFF)


def _update_node(ctx, node, node_info, valid_macs):
    if not node.maintenance:
        LOG.error(_LE('Refusing to apply discovered data to node %s '
                      'which is not in maintenance state'), node.uuid)
        return

    properties = node.properties.copy()
    for key in ('cpus', 'cpu_arch', 'memory_mb', 'local_gb'):
        if not properties.get(key):
            properties[key] = node_info[key]
    node.properties = properties
    node.extra = dict(node.extra, newly_discovered=True)
    node.save()

    for mac in valid_macs:
        try:
            new_port = objects.Port(ctx,
                                    node_id=node.id,
                                    address=mac)
            new_port.create()
        except exception.MACAlreadyExists:
            LOG.warning(_LW('MAC %(mac)s appeared in discovery data for '
                            'node %(node)s, but already exists in '
                            'database for another node - skipping'),
                        {'mac': mac, 'node': node.uuid})

    # Do it only after creating ports
    if not CONF.pxe.discovery_leave_maintenance:
        node.maintenance = False
        node.save()

    LOG.info(_LI('Node %s was updated with data from discovery process'),
             node.uuid)


_BULK_SIZE = 100


def _get_node_by_ipmi_address(ctx, address):
    marker = None
    condition = lambda n: n.driver_info.get('ipmi_address') != address
    while True:
        # NOTE(dtantsur): it's likely that the node was created recently,
        # so sort by created_at in reverse order
        node_list = objects.Node.list(ctx,
                                      limit=_BULK_SIZE,
                                      marker=marker,
                                      sort_key='created_at',
                                      sort_dir='desc')
        if not node_list:
            return

        node = next(itertools.dropwhile(condition, node_list), None)
        if node is not None:
            return node

        marker = node_list[-1]


def _get_node_by_macs(ctx, macs):
    port = None
    for mac in macs:
        try:
            port = objects.Port.get_by_address(ctx, mac)
        except exception.PortNotFound:
            continue
        else:
            break

    if port is None:
        return None
    return objects.Node.get_by_id(ctx, port.node_id)
