#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for the PXE discovery code.
"""

import mock

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as conductor_utils
from ironic.db import api as dbapi
from ironic.drivers.modules import pxe_discovery
from ironic import objects
from ironic.tests.db import base as db_base
from ironic.tests.objects import utils as obj_utils


DATA = {
    'ipmi_address': '1.2.3.4',
    'cpus': 2,
    'cpu_arch': 'x86_64',
    'memory_mb': 4096,
    'local_gb': 20,
    'macs': ['11:22:33:44:55:66',
             '66:55:44:33:22:11']
}


TEST_DRIVER = 'pxe_foobar'


@mock.patch.object(dbapi, 'get_instance')
class LockNodeTestCase(db_base.DbTestCase):

    def test_lock(self, mock_dbapi):
        with pxe_discovery._lock_node('node') as node:
            self.assertIs(mock_dbapi.return_value.reserve_node.return_value,
                          node)
        mock_dbapi.return_value.reserve_node.assert_called_once_with(mock.ANY,
                                                                     'node')
        mock_dbapi.return_value.release_node.assert_called_once_with(
            mock.ANY,
            mock_dbapi.return_value.reserve_node.return_value.id)


@mock.patch.object(driver_factory.DriverFactory, 'names', new=[TEST_DRIVER])
@mock.patch.object(pxe_discovery, '_update_node')
class ContinueDiscoveryTestCase(db_base.DbTestCase):

    def setUp(self):
        super(ContinueDiscoveryTestCase, self).setUp()
        self.config(discovery_enabled=True, group='pxe')
        self.config(discovery_allow_search_by_mac=True, group='pxe')

    def _test(self, **kwargs):
        self.data = dict(DATA, **kwargs)
        pxe_discovery.continue_discovery(self.context, self.data)

    def test_error(self, mock_update):
        self._test(error='boom!')
        self.assertFalse(mock_update.called)

    def test_missing(self, mock_update):
        self._test(macs=None)
        self.assertFalse(mock_update.called)

    def test_disabled(self, mock_update):
        self.config(discovery_enabled=False, group='pxe')
        self._test()
        self.assertFalse(mock_update.called)

    @mock.patch.object(objects.Node, 'list')
    def test_by_bmc_no_matching_node(self, mock_list, mock_update):
        node = obj_utils.get_test_node(self.context)
        mock_list.side_effect = ([node], [])
        self._test()
        mock_list.assert_any_call(mock.ANY,
                                  limit=mock.ANY,
                                  marker=None,
                                  sort_key='created_at',
                                  sort_dir='desc')
        mock_list.assert_any_call(mock.ANY,
                                  limit=mock.ANY,
                                  marker=node,
                                  sort_key='created_at',
                                  sort_dir='desc')
        self.assertFalse(mock_update.called)

    @mock.patch.object(objects.Port, 'get_by_address')
    def test_by_mac_no_matching_node(self, mock_get_by_addr, mock_update):
        mock_get_by_addr.side_effect = exception.PortNotFound(port='')
        self._test(ipmi_address=None)
        self.assertFalse(mock_update.called)

    @mock.patch.object(objects.Port, 'get_by_address')
    def test_by_mac_not_allowed(self, mock_get_by_addr, mock_update):
        self.config(discovery_allow_search_by_mac=False, group='pxe')
        self._test(ipmi_address=None)
        self.assertFalse(mock_get_by_addr.called)
        self.assertFalse(mock_update.called)

    @mock.patch.object(objects.Node, 'list')
    @mock.patch.object(pxe_discovery, '_lock_node')
    def test_by_bmc_ok(self, mock_lock, mock_list, mock_update):
        node = obj_utils.get_test_node(self.context, maintenance=True)
        node.driver_info['ipmi_address'] = DATA['ipmi_address']
        node.properties.clear()

        mock_lock.return_value.__enter__.return_value = node
        mock_list.side_effect = ([node], [])

        self._test()

        mock_list.assert_called_once_with(mock.ANY,
                                          limit=mock.ANY,
                                          marker=None,
                                          sort_key='created_at',
                                          sort_dir='desc')
        mock_lock.assert_called_once_with(node.uuid)
        mock_update.assert_called_once_with(self.context, node, self.data,
                                            self.data['macs'])

    @mock.patch.object(objects.Node, 'get_by_id')
    @mock.patch.object(objects.Port, 'get_by_address')
    @mock.patch.object(pxe_discovery, '_lock_node')
    def test_by_mac_ok(self, mock_lock, mock_get_port, mock_get_node,
                       mock_update):
        port = obj_utils.get_test_port(self.context)
        node = obj_utils.get_test_node(self.context, maintenance=True)
        node.properties.clear()

        mock_lock.return_value.__enter__.return_value = node
        mock_get_port.side_effect = (exception.PortNotFound(port=''), port)
        mock_get_node.return_value = node

        self._test(ipmi_address=None)

        mock_get_port.assert_any_call(mock.ANY, self.data['macs'][0])
        mock_get_port.assert_any_call(mock.ANY, self.data['macs'][1])
        mock_get_node.assert_called_once_with(mock.ANY, port.node_id)
        mock_lock.assert_called_once_with(node.uuid)
        mock_update.assert_called_once_with(self.context, node, self.data,
                                            self.data['macs'])

    @mock.patch.object(objects.Node, 'get_by_id')
    @mock.patch.object(objects.Port, 'get_by_address')
    @mock.patch.object(pxe_discovery, '_lock_node')
    def test_by_mac_filter_macs(self, mock_lock, mock_get_port,
                                mock_get_node, mock_update):
        port = obj_utils.get_test_port(self.context)
        node = obj_utils.get_test_node(self.context, maintenance=True)
        node.properties.clear()

        mock_lock.return_value.__enter__.return_value = node
        mock_get_port.return_value = port
        mock_get_node.return_value = node

        self._test(ipmi_address=None,
                   macs=['hello world', 'aa:BB:cc:DD:ee:FF'])

        mock_get_port.assert_called_once_with(mock.ANY, 'aa:bb:cc:dd:ee:ff')
        mock_get_node.assert_called_once_with(mock.ANY, port.node_id)
        mock_lock.assert_called_once_with(node.uuid)
        mock_update.assert_called_once_with(self.context, node, self.data,
                                            ['aa:bb:cc:dd:ee:ff'])

    @mock.patch.object(conductor_utils, 'node_power_action')
    @mock.patch.object(task_manager, 'acquire')
    @mock.patch.object(objects.Node, 'list')
    @mock.patch.object(pxe_discovery, '_lock_node')
    def test_power_off(self, mock_lock, mock_list,
                       mock_task_mgr, mock_power_action, mock_update):
        node = obj_utils.get_test_node(self.context, maintenance=True,
                                       driver=TEST_DRIVER)
        node.driver_info['ipmi_address'] = DATA['ipmi_address']
        node.properties.clear()

        mock_lock.return_value.__enter__.return_value = node
        mock_list.side_effect = ([node], [])

        self._test()

        mock_update.assert_called_once_with(self.context, node, self.data,
                                            self.data['macs'])
        mock_power_action.assert_called_once_with(
            mock_task_mgr.return_value.__enter__.return_value,
            states.POWER_OFF)


@mock.patch.object(objects, 'Port')
@mock.patch.object(objects.Node, 'save')
class UpdateNodeTestCase(db_base.DbTestCase):

    def _test(self, node, data=None):
        data = data or DATA
        pxe_discovery._update_node(self.context, node, data, data['macs'])

    def test_not_in_maintenance(self, mock_save, mock_port):
        node = obj_utils.get_test_node(self.context)
        self._test(node)
        self.assertFalse(mock_save.called)
        self.assertFalse(mock_port.called)

    def test_update_ok(self, mock_save, mock_port):
        node = obj_utils.get_test_node(self.context, maintenance=True)
        node.properties.clear()
        self._test(node)
        mock_save.assert_called_once_with()
        self.assertEqual({
            'cpus': 2,
            'cpu_arch': 'x86_64',
            'memory_mb': 4096,
            'local_gb': 20
        }, node.properties)
        self.assertTrue(node.maintenance)
        self.assertTrue(node.extra['newly_discovered'])
        self.assertEqual(2, mock_port.call_count)
        self.assertEqual(2, mock_port.return_value.create.call_count)
        mock_port.assert_called_with(self.context, node_id=node.id,
                                     address=mock.ANY)

    def test_update_mac_exists(self, mock_save, mock_port):
        node = obj_utils.get_test_node(self.context, maintenance=True)
        node.properties.clear()
        mock_port.side_effect = exception.MACAlreadyExists('error')
        self._test(node)
        mock_save.assert_called_once_with()
        self.assertEqual(2, mock_port.call_count)
        mock_port.assert_called_with(self.context, node_id=node.id,
                                     address=mock.ANY)

    def test_update_reset_maintenance(self, mock_save, mock_port):
        self.config(discovery_leave_maintenance=False, group='pxe')
        node = obj_utils.get_test_node(self.context, maintenance=True)
        node.properties.clear()
        self._test(node)
        # Called 2 times here
        mock_save.assert_called_with()
        self.assertEqual({
            'cpus': 2,
            'cpu_arch': 'x86_64',
            'memory_mb': 4096,
            'local_gb': 20
        }, node.properties)
        self.assertFalse(node.maintenance)
        self.assertTrue(node.extra['newly_discovered'])
        self.assertEqual(2, mock_port.call_count)

    def test_update_keeps_existing_properties(self, mock_save, mock_port):
        node = obj_utils.get_test_node(self.context, maintenance=True)
        old_properites = node.properties.copy()  # should not be updated
        self._test(node)
        mock_save.assert_called_once_with()
        self.assertEqual(old_properites, node.properties)
        self.assertTrue(node.maintenance)
        self.assertTrue(node.extra['newly_discovered'])
        self.assertEqual(2, mock_port.call_count)
